"""Basic tests for FluxFlow UI package."""


class TestUIImports:
    """Test that UI package imports work correctly."""

    def test_import_fluxflow_ui(self):
        """Test importing the main package."""
        import fluxflow_ui

        assert hasattr(fluxflow_ui, "__version__")

    def test_import_app_flask(self):
        """Test importing Flask app module."""
        from fluxflow_ui import app_flask

        assert app_flask is not None

    def test_import_config_manager(self):
        """Test importing config manager."""
        from fluxflow_ui.utils import config_manager

        assert config_manager is not None


class TestUIConfiguration:
    """Test UI configuration functionality."""

    def test_config_manager_exists(self):
        """Test that ConfigManager class exists."""
        from fluxflow_ui.utils.config_manager import ConfigManager

        assert ConfigManager is not None



"""Tests for Flask API endpoints."""

import json
from unittest.mock import patch

import pytest

from fluxflow_ui.app_flask import app


@pytest.fixture
def client():
    """Create test client."""
    app.config["TESTING"] = True
    with app.test_client() as client:
        yield client


class TestIndexEndpoint:
    """Tests for main page endpoint."""

    def test_index_returns_html(self, client):
        """Should return HTML page."""
        response = client.get("/")

        assert response.status_code == 200


class TestTrainingConfigEndpoints:
    """Tests for training config endpoints."""

    def test_get_training_config_returns_json(self, client):
        """GET should return JSON config."""
        response = client.get("/api/config/training")

        assert response.status_code == 200
        assert response.content_type == "application/json"
        data = json.loads(response.data)
        assert isinstance(data, dict)

    def test_post_training_config_requires_json(self, client):
        """POST should require JSON body."""
        response = client.post("/api/config/training")

        # 415 = UNSUPPORTED MEDIA TYPE (no Content-Type header)
        assert response.status_code in [400, 415]

    def test_post_training_config_saves(self, client):
        """POST should save config."""
        config = {"batch_size": 4}
        response = client.post(
            "/api/config/training", data=json.dumps(config), content_type="application/json"
        )

        assert response.status_code == 200
        data = json.loads(response.data)
        assert data["status"] == "success"


class TestGenerationConfigEndpoints:
    """Tests for generation config endpoints."""

    def test_get_generation_config_returns_json(self, client):
        """GET should return JSON config."""
        response = client.get("/api/config/generation")

        assert response.status_code == 200
        data = json.loads(response.data)
        assert isinstance(data, dict)

    def test_post_generation_config_requires_json(self, client):
        """POST should require JSON body."""
        response = client.post("/api/config/generation")

        assert response.status_code in [400, 415]

    def test_post_generation_config_saves(self, client):
        """POST should save config."""
        config = {"img_size": 512}
        response = client.post(
            "/api/config/generation", data=json.dumps(config), content_type="application/json"
        )

        assert response.status_code == 200


class TestTrainingEndpoints:
    """Tests for training control endpoints."""

    def test_start_training_requires_json(self, client):
        """POST start should require JSON body."""
        response = client.post("/api/training/start")

        assert response.status_code in [400, 415]

    @patch("fluxflow_ui.app_flask.training_runner")
    def test_start_training_returns_error_if_running(self, mock_runner, client):
        """Should return error if already running."""
        mock_runner.is_running = True

        response = client.post(
            "/api/training/start", data=json.dumps({}), content_type="application/json"
        )

        assert response.status_code == 400
        data = json.loads(response.data)
        # Check for either message format
        assert "already running" in data.get("message", "") or data.get("status") == "error"

    @patch("fluxflow_ui.app_flask.training_runner")
    def test_start_training_success(self, mock_runner, client):
        """Should start training successfully."""
        mock_runner.is_running = False
        mock_runner.start_training.return_value = True

        response = client.post(
            "/api/training/start",
            data=json.dumps({"batch_size": 4}),
            content_type="application/json",
        )

        assert response.status_code == 200
        data = json.loads(response.data)
        assert data["status"] == "success"

    @patch("fluxflow_ui.app_flask.training_runner")
    def test_stop_training_returns_error_if_not_running(self, mock_runner, client):
        """Should return error if not running."""
        mock_runner.is_running = False

        response = client.post("/api/training/stop")

        assert response.status_code == 400

    @patch("fluxflow_ui.app_flask.training_runner")
    def test_stop_training_success(self, mock_runner, client):
        """Should stop training successfully."""
        mock_runner.is_running = True
        mock_runner.stop_training.return_value = True

        response = client.post("/api/training/stop")

        assert response.status_code == 200

    @patch("fluxflow_ui.app_flask.training_runner")
    def test_training_status_returns_state(self, mock_runner, client):
        """Should return current training state."""
        mock_runner.is_running = True
        mock_runner.get_output.return_value = ["line 1", "line 2"]

        response = client.get("/api/training/status")

        assert response.status_code == 200
        data = json.loads(response.data)
        assert data["running"] is True
        assert data["output"] == ["line 1", "line 2"]


class TestGenerationEndpoints:
    """Tests for generation endpoints."""

    def test_inspect_model_requires_json(self, client):
        """Should require JSON body."""
        response = client.post("/api/generation/inspect")

        assert response.status_code in [400, 415]

    def test_inspect_model_requires_checkpoint_path(self, client):
        """Should require checkpoint_path."""
        response = client.post(
            "/api/generation/inspect", data=json.dumps({}), content_type="application/json"
        )

        assert response.status_code == 400
        data = json.loads(response.data)
        # Check message contains expected text or is an error status
        assert "checkpoint_path" in data.get("message", "") or data.get("status") == "error"

    def test_inspect_model_returns_error_for_invalid_path(self, client):
        """Should return error for non-existent path."""
        response = client.post(
            "/api/generation/inspect",
            data=json.dumps({"checkpoint_path": "/nonexistent/model.safetensors"}),
            content_type="application/json",
        )

        assert response.status_code == 400

    def test_load_model_requires_json(self, client):
        """Should require JSON body."""
        response = client.post("/api/generation/load")

        assert response.status_code in [400, 415]

    def test_load_model_requires_checkpoint_path(self, client):
        """Should require checkpoint_path."""
        response = client.post(
            "/api/generation/load", data=json.dumps({}), content_type="application/json"
        )

        assert response.status_code == 400

    def test_generate_image_requires_json(self, client):
        """Should require JSON body."""
        response = client.post("/api/generation/generate")

        assert response.status_code in [400, 415]

    def test_generate_image_requires_prompt(self, client):
        """Should require prompt."""
        response = client.post(
            "/api/generation/generate", data=json.dumps({}), content_type="application/json"
        )

        assert response.status_code == 400
        data = json.loads(response.data)
        assert "prompt" in data.get("message", "") or data.get("status") == "error"

    @patch("fluxflow_ui.app_flask.generation_worker")
    def test_generation_status_returns_state(self, mock_worker, client):
        """Should return generation worker state."""
        mock_worker.is_loaded.return_value = True
        mock_worker.config = {"vae_dim": 64}

        response = client.get("/api/generation/status")

        assert response.status_code == 200
        data = json.loads(response.data)
        assert data["loaded"] is True


class TestFileBrowserEndpoint:
    """Tests for file browser endpoint."""

    def test_browse_files_requires_json(self, client):
        """Should require JSON body."""
        response = client.post("/api/files/browse")

        assert response.status_code in [400, 415]

    def test_browse_files_returns_items(self, client, tmp_path):
        """Should return directory items."""
        # Create test files
        (tmp_path / "test.txt").touch()
        (tmp_path / "subdir").mkdir()

        response = client.post(
            "/api/files/browse",
            data=json.dumps({"path": str(tmp_path)}),
            content_type="application/json",
        )

        assert response.status_code == 200
        data = json.loads(response.data)
        assert data["status"] == "success"
        assert "items" in data

        names = [item["name"] for item in data["items"]]
        assert "test.txt" in names
        assert "subdir" in names

    def test_browse_files_defaults_to_cwd(self, client):
        """Should default to current directory."""
        response = client.post(
            "/api/files/browse", data=json.dumps({"path": "."}), content_type="application/json"
        )

        assert response.status_code == 200

    def test_browse_files_filters_by_type(self, client, tmp_path):
        """Should filter by file type."""
        (tmp_path / "model.safetensors").touch()
        (tmp_path / "other.txt").touch()

        response = client.post(
            "/api/files/browse",
            data=json.dumps({"path": str(tmp_path), "type": "safetensors"}),
            content_type="application/json",
        )

        assert response.status_code == 200
        data = json.loads(response.data)

        file_names = [item["name"] for item in data["items"] if item["type"] == "file"]
        assert "model.safetensors" in file_names
        assert "other.txt" not in file_names

    def test_browse_files_hides_dotfiles(self, client, tmp_path):
        """Should hide hidden files."""
        (tmp_path / ".hidden").touch()
        (tmp_path / "visible").touch()

        response = client.post(
            "/api/files/browse",
            data=json.dumps({"path": str(tmp_path)}),
            content_type="application/json",
        )

        assert response.status_code == 200
        data = json.loads(response.data)

        names = [item["name"] for item in data["items"]]
        assert ".hidden" not in names
        assert "visible" in names

    def test_browse_files_includes_parent_dir(self, client, tmp_path):
        """Should include parent directory option."""
        response = client.post(
            "/api/files/browse",
            data=json.dumps({"path": str(tmp_path)}),
            content_type="application/json",
        )

        assert response.status_code == 200
        data = json.loads(response.data)

        names = [item["name"] for item in data["items"]]
        assert ".." in names


class TestJSONValidation:
    """Tests for JSON validation decorator."""

    def test_empty_body_rejected(self, client):
        """Should reject empty request body."""
        response = client.post("/api/config/training", data="", content_type="application/json")

        assert response.status_code == 400

    def test_invalid_json_rejected(self, client):
        """Should reject invalid JSON."""
        response = client.post(
            "/api/config/training", data="not json", content_type="application/json"
        )

        # Flask returns 400 for invalid JSON
        assert response.status_code in [400, 415]

    def test_valid_json_accepted(self, client):
        """Should accept valid JSON."""
        response = client.post(
            "/api/config/training", data=json.dumps({"test": True}), content_type="application/json"
        )

        assert response.status_code == 200



"""Training tab for FluxFlow UI."""

from pathlib import Path
from typing import Tuple

import gradio as gr

from fluxflow_ui.utils.config_manager import ConfigManager
from fluxflow_ui.utils.training_runner import TrainingRunner


def create_training_tab(runner: TrainingRunner, config_mgr: ConfigManager) -> gr.Tab:
    """Create training interface tab.

    Args:
        runner: Training runner instance
        config_mgr: Configuration manager

    Returns:
        Gradio Tab component
    """

    output_lines: list[str] = []

    def start_training_handler(
        # Data
        data_path: str,
        captions_file: str,
        use_webdataset: bool,
        webdataset_token: str,
        webdataset_url: str,
        webdataset_image_key: str,
        webdataset_caption_key: str,
        # Model
        output_path: str,
        model_checkpoint: str,
        vae_dim: int,
        text_embedding_dim: int,
        feature_dim: int,
        feature_dim_disc: int,
        pretrained_bert_model: str,
        # Training
        n_epochs: int,
        batch_size: int,
        lr: float,
        lr_min: float,
        workers: int,
        preserve_lr: bool,
        optim_sched_config: str,
        training_steps: int,
        use_fp16: bool,
        initial_clipping_norm: float,
        # Advanced training
        use_gradient_checkpointing: bool,
        use_lpips: bool,
        lambda_lpips: float,
        use_cfg: bool,
        cfg_dropout_prob: float,
        use_multires: bool,
        reduced_min_sizes: str,
        # Training modes
        train_vae: bool,
        train_no_gan: bool,
        train_spade: bool,
        train_diff: bool,
        train_diff_full: bool,
        # KL divergence
        kl_beta: float,
        kl_warmup: int,
        kl_free_bits: float,
        # Output
        lambda_adv: float,
        sample_interval: int,
        log_interval: int,
        no_samples: bool,
        test_image_address: str,
        sample_captions: str,
        # Misc
        tokenizer_name: str,
        img_size: int,
        channels: int,
    ) -> Tuple[str, str, str]:
        """Handle training start.

        Returns:
            Tuple of (status, start button label, stop button label)
        """
        if runner.is_running:
            return "⚠️ Training is already running", "⏸️ Training...", "⏹️ Stop Training"

        # Validate inputs
        if not use_webdataset:
            if not data_path or not Path(data_path).exists():
                return "❌ Invalid data path", "▶️ Start Training", "⏹️ Stop Training"
            if not captions_file or not Path(captions_file).exists():
                return "❌ Invalid captions file", "▶️ Start Training", "⏹️ Stop Training"
        else:
            if not webdataset_token or webdataset_token == "your_token_here":
                return "❌ Invalid HuggingFace token", "▶️ Start Training", "⏹️ Stop Training"
            if not webdataset_url:
                return "❌ Invalid WebDataset URL", "▶️ Start Training", "⏹️ Stop Training"

        # Build config
        config = {
            # Data
            "data_path": data_path,
            "captions_file": captions_file,
            "use_webdataset": use_webdataset,
            "webdataset_token": webdataset_token if webdataset_token else None,
            "webdataset_url": webdataset_url if webdataset_url else None,
            "webdataset_image_key": webdataset_image_key if webdataset_image_key else "jpg",
            "webdataset_caption_key": (
                webdataset_caption_key if webdataset_caption_key else "prompt"
            ),
            # Model
            "output_path": output_path,
            "model_checkpoint": model_checkpoint if model_checkpoint else None,
            "vae_dim": int(vae_dim),
            "text_embedding_dim": int(text_embedding_dim),
            "feature_maps_dim": int(feature_dim),
            "feature_maps_dim_disc": int(feature_dim_disc),
            "pretrained_bert_model": pretrained_bert_model if pretrained_bert_model else None,
            # Training
            "n_epochs": int(n_epochs),
            "batch_size": int(batch_size),
            "lr": float(lr),
            "lr_min": float(lr_min),
            "workers": int(workers),
            "preserve_lr": preserve_lr,
            "optim_sched_config": optim_sched_config if optim_sched_config else None,
            "training_steps": int(training_steps),
            "use_fp16": use_fp16,
            "initial_clipping_norm": float(initial_clipping_norm),
            # Advanced training
            "use_gradient_checkpointing": use_gradient_checkpointing,
            "use_lpips": use_lpips,
            "lambda_lpips": float(lambda_lpips),
            "cfg_dropout_prob": float(cfg_dropout_prob) if use_cfg else 0.0,
            "reduced_min_sizes": reduced_min_sizes if use_multires else "",
            # Training modes
            "train_vae": train_vae,
            "train_no_gan": train_no_gan,
            "train_spade": train_spade,
            "train_diff": train_diff,
            "train_diff_full": train_diff_full,
            # KL divergence
            "kl_beta": float(kl_beta),
            "kl_warmup_steps": int(kl_warmup),
            "kl_free_bits": float(kl_free_bits),
            # Output
            "lambda_adv": float(lambda_adv),
            "sample_interval": int(sample_interval),
            "log_interval": int(log_interval),
            "no_samples": no_samples,
            "test_image_address": test_image_address if test_image_address else None,
            "sample_captions": sample_captions if sample_captions else None,
            # Misc
            "tokenizer_name": tokenizer_name,
            "img_size": int(img_size),
            "channels": int(channels),
        }

        # Save config
        config_mgr.save_training_config(config)

        # Start training
        success = runner.start_training(config)

        if success:
            output_lines.clear()
            return "✅ Training started!", "⏸️ Training...", "⏹️ Stop Training"
        else:
            return "❌ Failed to start training", "▶️ Start Training", "⏹️ Stop Training"

    def stop_training_handler() -> Tuple[str, str, str]:
        """Handle training stop.

        Returns:
            Tuple of (status, start button label, stop button label)
        """
        if not runner.is_running:
            return "⚠️ No training is running", "▶️ Start Training", "⏹️ Stop Training"

        success = runner.stop_training()

        if success:
            return "✅ Training stopped", "▶️ Start Training", "⏹️ Stop Training"
        else:
            return "❌ Failed to stop training", "⏸️ Training...", "⏹️ Stop Training"

    def update_output() -> str:
        """Update console output.

        Returns:
            Console output text
        """
        new_lines = runner.get_output()
        output_lines.extend(new_lines)

        # Keep last 500 lines
        if len(output_lines) > 500:
            output_lines[:] = output_lines[-500:]

        return "\n".join(output_lines)

    # Load previous config
    prev_config = config_mgr.load_training_config()
    if not prev_config:
        prev_config = config_mgr.get_default_training_config()

    tab: gr.Tab
    with gr.Tab("🚀 Train") as tab:
        gr.Markdown("# FluxFlow Model Training")
        gr.Markdown(
            "Configure and train FluxFlow text-to-image models. "
            "See TRAINING_GUIDE.md for detailed documentation."
        )

        with gr.Row():
            # Left column: Configuration
            with gr.Column(scale=1):
                gr.Markdown("### Dataset Configuration")

                use_webdataset_checkbox = gr.Checkbox(
                    label="Use WebDataset Streaming",
                    value=prev_config.get("use_webdataset", False),
                    info="Stream from HuggingFace datasets (no download needed)",
                )

                with gr.Column(
                    visible=not prev_config.get("use_webdataset", False)
                ) as local_data_group:
                    data_path_input = gr.Textbox(
                        label="Data Path",
                        placeholder="/path/to/images",
                        value=prev_config.get("data_path", ""),
                    )
                    captions_file_input = gr.Textbox(
                        label="Captions File",
                        placeholder="/path/to/captions.txt",
                        value=prev_config.get("captions_file", ""),
                    )

                with gr.Column(
                    visible=prev_config.get("use_webdataset", False)
                ) as webdataset_data_group:
                    webdataset_token_input = gr.Textbox(
                        label="HuggingFace Token",
                        placeholder="hf_your_token_here",
                        value=prev_config.get("webdataset_token", ""),
                        type="password",
                    )
                    webdataset_url_input = gr.Textbox(
                        label="WebDataset URL Pattern",
                        placeholder="hf://datasets/username/dataset/*.tar",
                        value=prev_config.get("webdataset_url", ""),
                        info="HuggingFace dataset URL with tar pattern",
                    )
                    with gr.Row():
                        webdataset_image_key_input = gr.Textbox(
                            label="Image Key",
                            value=prev_config.get("webdataset_image_key", "jpg"),
                            info="File extension in tar (e.g., jpg, png)",
                        )
                        webdataset_caption_key_input = gr.Textbox(
                            label="Caption Key",
                            value=prev_config.get("webdataset_caption_key", "prompt"),
                            info="JSON field for captions",
                        )

                output_path_input = gr.Textbox(
                    label="Output Path",
                    value=prev_config.get("output_path", "outputs/flux"),
                )
                model_checkpoint_input = gr.Textbox(
                    label="Resume from Checkpoint (optional)",
                    placeholder="outputs/flux/flxflow_final.safetensors",
                    value=prev_config.get("model_checkpoint", ""),
                )

                gr.Markdown("### Model Architecture")

                with gr.Row():
                    vae_dim_input = gr.Number(
                        label="VAE Dimension",
                        value=prev_config.get("vae_dim", 64),
                        precision=0,
                        info="32=Low VRAM, 64=Balanced, 128=High Quality, 256=Max",
                    )
                    text_embedding_dim_input = gr.Number(
                        label="Text Embedding Dim",
                        value=prev_config.get("text_embedding_dim", 1024),
                        precision=0,
                    )

                with gr.Row():
                    feature_dim_input = gr.Number(
                        label="Feature Dimension",
                        value=prev_config.get("feature_maps_dim", 64),
                        precision=0,
                    )
                    feature_dim_disc_input = gr.Number(
                        label="Discriminator Dim",
                        value=prev_config.get("feature_maps_dim_disc", 128),
                        precision=0,
                    )

                pretrained_bert_input = gr.Textbox(
                    label="Pretrained BERT (optional)",
                    placeholder="path/to/bert_checkpoint.safetensors",
                    value=prev_config.get("pretrained_bert_model", ""),
                )

                gr.Markdown("### Training Parameters")

                with gr.Row():
                    n_epochs_input = gr.Number(
                        label="Epochs",
                        value=prev_config.get("n_epochs", 10),
                        precision=0,
                    )
                    batch_size_input = gr.Number(
                        label="Batch Size",
                        value=prev_config.get("batch_size", 1),
                        precision=0,
                    )

                with gr.Row():
                    lr_input = gr.Number(
                        label="Learning Rate",
                        value=prev_config.get("lr", 1e-5),
                        step=1e-6,
                        info="VAE: 1e-5, Flow: 5e-7",
                    )
                    lr_min_input = gr.Number(
                        label="Min LR Multiplier",
                        value=prev_config.get("lr_min", 0.1),
                        step=0.01,
                    )

                with gr.Row():
                    workers_input = gr.Number(
                        label="Workers",
                        value=prev_config.get("workers", 4),
                        precision=0,
                    )
                    training_steps_input = gr.Number(
                        label="Training Steps (Grad Accum)",
                        value=prev_config.get("training_steps", 1),
                        precision=0,
                        info="Effective batch = batch_size * training_steps",
                    )

                with gr.Row():
                    initial_clipping_norm_input = gr.Number(
                        label="Gradient Clipping Norm",
                        value=prev_config.get("initial_clipping_norm", 1.0),
                        step=0.1,
                    )
                    use_fp16_checkbox = gr.Checkbox(
                        label="Use FP16 (Mixed Precision)",
                        value=prev_config.get("use_fp16", False),
                        info="~40% faster on RTX GPUs",
                    )
                    preserve_lr_checkbox = gr.Checkbox(
                        label="Preserve LR",
                        value=prev_config.get("preserve_lr", False),
                    )

                optim_sched_config_input = gr.Textbox(
                    label="Optimizer/Scheduler Config (optional)",
                    placeholder="path/to/optim_config.json",
                    value=prev_config.get("optim_sched_config", ""),
                    info="Advanced: JSON file for per-model optimizer/scheduler config",
                )

                gr.Markdown("### Training Modes")

                with gr.Row():
                    train_vae_checkbox = gr.Checkbox(
                        label="Train VAE",
                        value=prev_config.get("train_vae", True),
                    )
                    train_no_gan_checkbox = gr.Checkbox(
                        label="Disable GAN",
                        value=prev_config.get("train_no_gan", False),
                    )
                    train_spade_checkbox = gr.Checkbox(
                        label="Use SPADE",
                        value=prev_config.get("train_spade", True),
                    )

                with gr.Row():
                    train_diff_checkbox = gr.Checkbox(
                        label="Train Flow",
                        value=prev_config.get("train_diff", False),
                    )
                    train_diff_full_checkbox = gr.Checkbox(
                        label="Train Flow (Full)",
                        value=prev_config.get("train_diff_full", False),
                    )

                with gr.Accordion("KL Divergence Settings", open=False):
                    kl_beta_input = gr.Slider(
                        label="KL Beta",
                        minimum=0.0,
                        maximum=5.0,
                        step=0.0001,
                        value=prev_config.get("kl_beta", 1.0),
                        info="Regularization strength",
                    )
                    kl_warmup_input = gr.Number(
                        label="KL Warmup Steps",
                        value=prev_config.get("kl_warmup_steps", 100000),
                        precision=0,
                    )
                    kl_free_bits_input = gr.Number(
                        label="KL Free Bits (nats)",
                        value=prev_config.get("kl_free_bits", 0.0),
                        step=0.1,
                        info="Minimum KL before penalty",
                    )

                with gr.Accordion("Output & Sampling Settings", open=False):
                    lambda_adv_input = gr.Slider(
                        label="Adversarial Weight",
                        minimum=0.0,
                        maximum=2.0,
                        step=0.1,
                        value=prev_config.get("lambda_adv", 0.9),
                    )
                    sample_interval_input = gr.Number(
                        label="Sample Interval (batches)",
                        value=prev_config.get("sample_interval", 50),
                        precision=0,
                    )
                    log_interval_input = gr.Number(
                        label="Log Interval (batches)",
                        value=prev_config.get("log_interval", 10),
                        precision=0,
                    )
                    no_samples_checkbox = gr.Checkbox(
                        label="Disable Sampling (faster training)",
                        value=prev_config.get("no_samples", False),
                    )
                    test_image_address_input = gr.Textbox(
                        label="Test Images for VAE (space-separated paths)",
                        placeholder="test1.jpg test2.png",
                        value=prev_config.get("test_image_address", ""),
                    )
                    sample_captions_input = gr.Textbox(
                        label="Sample Captions for Flow (one per line)",
                        placeholder="a photo of a cat\nan illustration of mountains",
                        value=prev_config.get("sample_captions", "a photo of a cat"),
                        lines=3,
                    )

                with gr.Accordion("Advanced Training Parameters", open=False):
                    with gr.Row():
                        use_gradient_checkpointing_checkbox = gr.Checkbox(
                            label="Gradient Checkpointing",
                            value=prev_config.get("use_gradient_checkpointing", False),
                            info="Save VRAM at cost of speed",
                        )
                        use_lpips_checkbox = gr.Checkbox(
                            label="Use LPIPS Perceptual Loss",
                            value=prev_config.get("use_lpips", True),
                            info="Improves perceptual quality (adds ~2GB VRAM)",
                        )

                    lambda_lpips_input = gr.Slider(
                        label="LPIPS Weight",
                        minimum=0.0,
                        maximum=1.0,
                        step=0.05,
                        value=prev_config.get("lambda_lpips", 0.1),
                        info="Perceptual loss strength (0.1 recommended)",
                    )

                with gr.Accordion("Classifier-Free Guidance (CFG)", open=False):
                    gr.Markdown(
                        """
                        **CFG Training** enables guided generation at inference time.
                        Train with CFG dropout to support guidance_scale > 1.0.
                        """
                    )
                    use_cfg_checkbox = gr.Checkbox(
                        label="Enable CFG Training",
                        value=prev_config.get("cfg_dropout_prob", 0.0) > 0.0,
                        info="Required for guided generation",
                    )
                    cfg_dropout_prob_input = gr.Slider(
                        label="CFG Dropout Probability",
                        minimum=0.0,
                        maximum=0.20,
                        step=0.01,
                        value=prev_config.get("cfg_dropout_prob", 0.10),
                        info="Recommended: 0.10 (10% null conditioning)",
                        visible=prev_config.get("cfg_dropout_prob", 0.0) > 0.0,
                    )

                    def toggle_cfg(enabled):
                        return gr.update(visible=enabled)

                    use_cfg_checkbox.change(
                        fn=toggle_cfg,
                        inputs=[use_cfg_checkbox],
                        outputs=[cfg_dropout_prob_input],
                    )

                with gr.Accordion("Multi-Resolution Training", open=False):
                    gr.Markdown(
                        """
                        **Progressive resolution** trains on smaller images first,
                        then gradually increases. Improves convergence and reduces
                        initial VRAM usage.
                        """
                    )
                    use_multires_checkbox = gr.Checkbox(
                        label="Enable Progressive Resolution",
                        value=bool(prev_config.get("reduced_min_sizes", "")),
                    )
                    reduced_min_sizes_input = gr.Textbox(
                        label="Resolution Stages (comma-separated)",
                        placeholder="256, 384, 512, 768, 1024",
                        value=prev_config.get("reduced_min_sizes", ""),
                        info="Trains progressively from smallest to img_size",
                        visible=bool(prev_config.get("reduced_min_sizes", "")),
                    )

                    def toggle_multires(enabled):
                        return gr.update(visible=enabled)

                    use_multires_checkbox.change(
                        fn=toggle_multires,
                        inputs=[use_multires_checkbox],
                        outputs=[reduced_min_sizes_input],
                    )

                with gr.Accordion("Misc Settings", open=False):
                    tokenizer_name_input = gr.Textbox(
                        label="Tokenizer Name",
                        value=prev_config.get("tokenizer_name", "distilbert-base-uncased"),
                    )
                    img_size_input = gr.Number(
                        label="Image Size (px)",
                        value=prev_config.get("img_size", 1024),
                        precision=0,
                        info="Images resized to this size",
                    )
                    channels_input = gr.Number(
                        label="Image Channels",
                        value=prev_config.get("channels", 3),
                        precision=0,
                        info="3=RGB, 1=Grayscale",
                    )

                with gr.Row():
                    start_btn = gr.Button("▶️ Start Training", variant="primary", scale=3)
                    stop_btn = gr.Button("⏹️ Stop Training", variant="stop", scale=1)

                status_output = gr.Textbox(
                    label="Status",
                    value="Ready to train",
                    interactive=False,
                )

            # Right column: Output
            with gr.Column(scale=1):
                gr.Markdown("### Training Console")
                console_output = gr.Textbox(
                    label="Output",
                    lines=30,
                    max_lines=30,
                    interactive=False,
                    autoscroll=True,
                )

                gr.Markdown("### Quick Guide")
                gr.Markdown(
                    """
                **Training Stages:**
                1. **VAE Pretraining** (50-100 epochs)
                   - Enable: Train VAE, Use SPADE
                   - LR: 1e-5, Disable: Train Flow
                2. **Flow Training** (100-200 epochs)
                   - Enable: Train Flow (Full)
                   - LR: 5e-7, Load VAE checkpoint
                3. **Joint Fine-tuning** (20-50 epochs, optional)
                   - Enable: Train VAE, Train Flow (Full), Use SPADE
                   - LR: 1e-6

                **VRAM Settings:**
                - **8GB**: dim=32, batch=1, fp16=on, img_size=512
                - **12-16GB**: dim=64, batch=2, fp16=on
                - **24GB+**: dim=128, batch=4, fp16=on

                **Advanced Features:**
                - **Optimizer/Scheduler Config**: Use custom optimizers (Lion, AdamW, etc.)
                  and schedulers per model via JSON file
                - See TRAINING_GUIDE.md for complete documentation and examples

                **Tips:**
                - Monitor console for loss values
                - Use FP16 for 2-4x speedup on RTX GPUs
                - TTI-2M streams 2M+ images (no download needed)
                - Default optimizers: Lion (flow), AdamW (vae/discriminator)
                """
                )

        # Toggle dataset inputs based on use_webdataset
        def toggle_dataset_inputs(use_webdataset):
            return gr.update(visible=not use_webdataset), gr.update(visible=use_webdataset)

        use_webdataset_checkbox.change(
            fn=toggle_dataset_inputs,
            inputs=[use_webdataset_checkbox],
            outputs=[local_data_group, webdataset_data_group],
        )

        # Event handlers
        start_btn.click(
            fn=start_training_handler,
            inputs=[
                # Data
                data_path_input,
                captions_file_input,
                use_webdataset_checkbox,
                webdataset_token_input,
                webdataset_url_input,
                webdataset_image_key_input,
                webdataset_caption_key_input,
                # Model
                output_path_input,
                model_checkpoint_input,
                vae_dim_input,
                text_embedding_dim_input,
                feature_dim_input,
                feature_dim_disc_input,
                pretrained_bert_input,
                # Training
                n_epochs_input,
                batch_size_input,
                lr_input,
                lr_min_input,
                workers_input,
                preserve_lr_checkbox,
                optim_sched_config_input,
                training_steps_input,
                use_fp16_checkbox,
                initial_clipping_norm_input,
                # Advanced training
                use_gradient_checkpointing_checkbox,
                use_lpips_checkbox,
                lambda_lpips_input,
                use_cfg_checkbox,
                cfg_dropout_prob_input,
                use_multires_checkbox,
                reduced_min_sizes_input,
                # Training modes
                train_vae_checkbox,
                train_no_gan_checkbox,
                train_spade_checkbox,
                train_diff_checkbox,
                train_diff_full_checkbox,
                # KL divergence
                kl_beta_input,
                kl_warmup_input,
                kl_free_bits_input,
                # Output
                lambda_adv_input,
                sample_interval_input,
                log_interval_input,
                no_samples_checkbox,
                test_image_address_input,
                sample_captions_input,
                # Misc
                tokenizer_name_input,
                img_size_input,
                channels_input,
            ],
            outputs=[status_output, start_btn, stop_btn],
        )

        stop_btn.click(
            fn=stop_training_handler,
            outputs=[status_output, start_btn, stop_btn],
        )

        # Auto-update console every 1 second
        console_output.change(
            fn=lambda: None,
            outputs=[],
            every=1,
        )

        demo_timer = gr.Timer(value=1, active=True)
        demo_timer.tick(
            fn=update_output,
            outputs=[console_output],
        )

    return tab



"""Configuration management for UI."""

import json
import logging
from pathlib import Path
from typing import Any, Dict, Optional

logger = logging.getLogger(__name__)


class ConfigManager:
    """Manage training and generation configurations."""

    def __init__(self, config_dir: str = ".ui_configs"):
        """Initialize config manager.

        Args:
            config_dir: Directory to store config files
        """
        self.config_dir = Path(config_dir)
        self.config_dir.mkdir(exist_ok=True)
        self.training_config_path = self.config_dir / "training_config.json"
        self.generation_config_path = self.config_dir / "generation_config.json"

    def save_training_config(self, config: Dict[str, Any]) -> bool:
        """Save training configuration.

        Args:
            config: Training configuration dictionary

        Returns:
            True if saved successfully, False otherwise
        """
        try:
            with open(self.training_config_path, "w") as f:
                json.dump(config, f, indent=2)
            return True
        except (IOError, OSError) as e:
            logger.error(f"Failed to save training config: {e}")
            return False
        except (TypeError, ValueError) as e:
            logger.error(f"Invalid training config data: {e}")
            return False

    def load_training_config(self) -> Optional[Dict[str, Any]]:
        """Load training configuration.

        Returns:
            Training configuration dict or None if not found/invalid
        """
        if not self.training_config_path.exists():
            return None
        try:
            with open(self.training_config_path, "r") as f:
                data: Dict[str, Any] = json.load(f)
                return data
        except json.JSONDecodeError as e:
            logger.error(f"Corrupted training config file: {e}")
            return None
        except (IOError, OSError) as e:
            logger.error(f"Failed to read training config: {e}")
            return None

    def save_generation_config(self, config: Dict[str, Any]) -> bool:
        """Save generation configuration.

        Args:
            config: Generation configuration dictionary

        Returns:
            True if saved successfully, False otherwise
        """
        try:
            with open(self.generation_config_path, "w") as f:
                json.dump(config, f, indent=2)
            return True
        except (IOError, OSError) as e:
            logger.error(f"Failed to save generation config: {e}")
            return False
        except (TypeError, ValueError) as e:
            logger.error(f"Invalid generation config data: {e}")
            return False

    def load_generation_config(self) -> Optional[Dict[str, Any]]:
        """Load generation configuration.

        Returns:
            Generation configuration dict or None if not found/invalid
        """
        if not self.generation_config_path.exists():
            return None
        try:
            with open(self.generation_config_path, "r") as f:
                data: Dict[str, Any] = json.load(f)
                return data
        except json.JSONDecodeError as e:
            logger.error(f"Corrupted generation config file: {e}")
            return None
        except (IOError, OSError) as e:
            logger.error(f"Failed to read generation config: {e}")
            return None

    def get_default_training_config(self) -> Dict[str, Any]:
        """Get default training configuration.

        Returns:
            Default training configuration
        """
        return {
            # Data
            "data_path": "",
            "captions_file": "",
            "use_webdataset": False,
            "webdataset_token": "",
            "webdataset_url": "",
            "webdataset_image_key": "jpg",
            "webdataset_caption_key": "prompt",
            # Legacy (deprecated but kept for compatibility)
            "use_tt2m": False,
            "tt2m_token": "",
            # Model
            "output_path": "outputs/flux",
            "model_checkpoint": "",
            "vae_dim": 64,
            "text_embedding_dim": 1024,
            "feature_maps_dim": 64,
            "feature_maps_dim_disc": 128,
            "pretrained_bert_model": "",
            # Training
            "n_epochs": 10,
            "batch_size": 1,
            "lr": 1e-5,
            "lr_min": 0.1,
            "workers": 4,
            "preserve_lr": False,
            "optim_sched_config": "",
            "training_steps": 1,
            "use_fp16": False,
            "initial_clipping_norm": 1.0,
            # Advanced training
            "use_gradient_checkpointing": False,
            "use_lpips": True,
            "lambda_lpips": 0.1,
            "cfg_dropout_prob": 0.0,
            "reduced_min_sizes": "",
            # Training modes
            "train_vae": True,
            "train_no_gan": False,
            "train_spade": True,
            "train_diff": False,
            "train_diff_full": False,
            # KL divergence
            "kl_beta": 1.0,
            "kl_warmup_steps": 100000,
            "kl_free_bits": 0.0,
            # Output
            "lambda_adv": 0.9,
            "sample_interval": 50,
            "log_interval": 10,
            "no_samples": False,
            "test_image_address": "",
            "sample_captions": "a photo of a cat",
            # Misc
            "tokenizer_name": "distilbert-base-uncased",
            "img_size": 1024,
            "channels": 3,
            # Optimizer/Scheduler configurations
            "optimizers": {
                "flow": {
                    "type": "Lion",
                    "lr": 5e-7,
                    "betas": [0.9, 0.95],
                    "weight_decay": 0.01,
                    "decoupled_weight_decay": True,
                },
                "vae": {
                    "type": "AdamW",
                    "lr": 5e-7,
                    "betas": [0.9, 0.95],
                    "weight_decay": 0.01,
                },
                "text_encoder": {
                    "type": "AdamW",
                    "lr": 5e-8,
                    "betas": [0.9, 0.99],
                    "weight_decay": 0.01,
                },
                "discriminator": {
                    "type": "AdamW",
                    "lr": 5e-7,
                    "betas": [0.0, 0.9],
                    "weight_decay": 0.001,
                    "amsgrad": True,
                },
            },
            "schedulers": {
                "flow": {
                    "type": "CosineAnnealingLR",
                    "eta_min_factor": 0.1,
                },
                "vae": {
                    "type": "CosineAnnealingLR",
                    "eta_min_factor": 0.1,
                },
                "text_encoder": {
                    "type": "CosineAnnealingLR",
                    "eta_min_factor": 0.001,
                },
                "discriminator": {
                    "type": "CosineAnnealingLR",
                    "eta_min_factor": 0.1,
                },
            },
        }

    def get_default_generation_config(self) -> Dict[str, Any]:
        """Get default generation configuration.

        Returns:
            Default generation configuration
        """
        return {
            "model_checkpoint": "",
            "output_path": "generated",
            "img_size": 512,
            "ddim_steps": 50,
            "batch_size": 1,
            "vae_dim": 64,
            "feature_maps_dim": 64,
            "text_embedding_dim": 1024,
            # CFG parameters
            "use_cfg": False,
            "guidance_scale": 5.0,
            "negative_prompt": "",
        }



"""Generation tab for FluxFlow UI."""

from pathlib import Path
from typing import Optional, Tuple

import gradio as gr
import numpy as np

from fluxflow_ui.utils.config_manager import ConfigManager
from fluxflow_ui.utils.generation_worker import GenerationWorker


def create_generation_tab(worker: GenerationWorker, config_mgr: ConfigManager) -> gr.Tab:
    """Create generation interface tab.

    Args:
        worker: Generation worker instance
        config_mgr: Configuration manager

    Returns:
        Gradio Tab component
    """

    def load_model_handler(
        checkpoint: str,
        vae_dim: int,
        feature_dim: int,
        text_dim: int,
    ) -> Tuple[str, str]:
        """Handle model loading.

        Returns:
            Tuple of (status message, button label)
        """
        if not checkpoint or not Path(checkpoint).exists():
            return "❌ Please provide a valid checkpoint path", "🔄 Load Model"

        success, message = worker.load_model(
            checkpoint_path=checkpoint,
            vae_dim=vae_dim,
            feature_maps_dim=feature_dim,
            text_embedding_dim=text_dim,
        )

        if success:
            config_mgr.save_generation_config(
                {
                    "model_checkpoint": checkpoint,
                    "vae_dim": vae_dim,
                    "feature_maps_dim": feature_dim,
                    "text_embedding_dim": text_dim,
                }
            )
            return f"✅ {message}", "✅ Model Loaded"
        else:
            return f"❌ {message}", "🔄 Load Model"

    def generate_handler(
        prompt: str,
        img_size: int,
        steps: int,
        seed: Optional[int],
        use_seed: bool,
        use_cfg: bool,
        guidance_scale: float,
        negative_prompt: str,
    ) -> Tuple[Optional[np.ndarray], str]:
        """Handle image generation.

        Returns:
            Tuple of (image, status message)
        """
        if not worker.is_loaded():
            return None, "❌ Please load a model first"

        if not prompt.strip():
            return None, "❌ Please enter a prompt"

        actual_seed = seed if use_seed else None
        image, message = worker.generate_image(
            prompt=prompt,
            img_width=img_size,
            img_height=img_size,
            ddim_steps=steps,
            seed=actual_seed,
            use_cfg=use_cfg,
            guidance_scale=guidance_scale if use_cfg else 1.0,
            negative_prompt=negative_prompt if use_cfg and negative_prompt.strip() else None,
        )

        return image, message

    # Load previous config
    prev_config = config_mgr.load_generation_config()
    if not prev_config:
        prev_config = config_mgr.get_default_generation_config()

    tab: gr.Tab
    with gr.Tab("🎨 Generate") as tab:
        gr.Markdown("# FluxFlow Image Generation")
        gr.Markdown("Generate images from text prompts using trained models")

        with gr.Row():
            # Left column: Configuration
            with gr.Column(scale=1):
                gr.Markdown("### Model Configuration")

                checkpoint_input = gr.Textbox(
                    label="Model Checkpoint",
                    placeholder="outputs/flux/flxflow_final.safetensors",
                    value=prev_config.get("model_checkpoint", ""),
                )

                with gr.Row():
                    vae_dim_input = gr.Number(
                        label="VAE Dimension",
                        value=prev_config.get("vae_dim", 64),
                        precision=0,
                    )
                    feature_dim_input = gr.Number(
                        label="Feature Dimension",
                        value=prev_config.get("feature_maps_dim", 64),
                        precision=0,
                    )

                text_dim_input = gr.Number(
                    label="Text Embedding Dimension",
                    value=prev_config.get("text_embedding_dim", 1024),
                    precision=0,
                )

                load_btn = gr.Button("🔄 Load Model", variant="primary")
                load_status = gr.Textbox(
                    label="Status",
                    value="Model not loaded",
                    interactive=False,
                )

                gr.Markdown("---")
                gr.Markdown("### Generation Parameters")

                prompt_input = gr.Textbox(
                    label="Prompt",
                    placeholder="A beautiful sunset over mountains",
                    lines=3,
                )

                with gr.Row():
                    img_size_input = gr.Slider(
                        label="Image Size",
                        minimum=256,
                        maximum=1024,
                        step=64,
                        value=512,
                    )
                    steps_input = gr.Slider(
                        label="Sampling Steps",
                        minimum=10,
                        maximum=100,
                        step=5,
                        value=50,
                    )

                with gr.Row():
                    use_seed_checkbox = gr.Checkbox(
                        label="Use Fixed Seed",
                        value=False,
                    )
                    seed_input = gr.Number(
                        label="Seed",
                        value=42,
                        precision=0,
                    )

                with gr.Accordion("Classifier-Free Guidance", open=False):
                    gr.Markdown(
                        """
                        **CFG** increases prompt adherence. Requires model trained with CFG dropout.
                        """
                    )
                    use_cfg_checkbox = gr.Checkbox(
                        label="Enable CFG",
                        value=False,
                        info="Model must be trained with cfg_dropout_prob > 0",
                    )
                    guidance_scale_input = gr.Slider(
                        label="Guidance Scale",
                        minimum=1.0,
                        maximum=15.0,
                        step=0.5,
                        value=5.0,
                        info="Higher = stronger prompt adherence (3-7 recommended)",
                        visible=False,
                    )
                    negative_prompt_input = gr.Textbox(
                        label="Negative Prompt (Optional)",
                        placeholder="blurry, low quality, distorted",
                        lines=2,
                        info="What to avoid in the image",
                        visible=False,
                    )

                    def toggle_cfg_gen(enabled):
                        return gr.update(visible=enabled), gr.update(visible=enabled)

                    use_cfg_checkbox.change(
                        fn=toggle_cfg_gen,
                        inputs=[use_cfg_checkbox],
                        outputs=[guidance_scale_input, negative_prompt_input],
                    )

                generate_btn = gr.Button("✨ Generate Image", variant="primary", size="lg")
                gen_status = gr.Textbox(
                    label="Generation Status",
                    value="Ready to generate",
                    interactive=False,
                )

            # Right column: Output
            with gr.Column(scale=1):
                gr.Markdown("### Generated Image")
                output_image = gr.Image(
                    label="Result",
                    type="numpy",
                    height=600,
                )

                gr.Markdown("### Tips")
                gr.Markdown(
                    """
                - **Model Loading**: Select a checkpoint and click Load Model first
                - **Prompt**: Describe the image you want to generate
                - **Image Size**: Larger sizes take longer but produce more detail
                - **Steps**: More steps = higher quality but slower (50 is usually good)
                - **Seed**: Enable for reproducible results
                """
                )

        # Event handlers
        load_btn.click(
            fn=load_model_handler,
            inputs=[checkpoint_input, vae_dim_input, feature_dim_input, text_dim_input],
            outputs=[load_status, load_btn],
        )

        generate_btn.click(
            fn=generate_handler,
            inputs=[
                prompt_input,
                img_size_input,
                steps_input,
                seed_input,
                use_seed_checkbox,
                use_cfg_checkbox,
                guidance_scale_input,
                negative_prompt_input,
            ],
            outputs=[output_image, gen_status],
        )

    return tab



"""Tests for TrainingRunner."""

import queue
import threading
from unittest.mock import MagicMock, patch

from fluxflow_ui.utils.training_runner import TrainingRunner


class TestTrainingRunnerInit:
    """Tests for TrainingRunner initialization."""

    def test_initializes_with_defaults(self):
        """Should initialize with correct default values."""
        runner = TrainingRunner()

        assert runner.process is None
        assert runner.is_running is False
        assert runner.thread is None
        assert runner.temp_config_file is None
        assert isinstance(runner.output_queue, queue.Queue)

    def test_is_running_is_thread_safe(self):
        """is_running property should be thread-safe."""
        runner = TrainingRunner()

        # Should be accessible without errors
        assert runner.is_running is False


class TestStartTraining:
    """Tests for starting training."""

    def test_returns_false_if_already_running(self):
        """Should return False if training is already running."""
        runner = TrainingRunner()
        runner._is_running = True

        result = runner.start_training({"data_path": "/test"})

        assert result is False

    @patch("subprocess.Popen")
    def test_starts_process_with_correct_command(self, mock_popen):
        """Should start subprocess with fluxflow-train command."""
        mock_process = MagicMock()
        mock_process.stdout = iter([])
        mock_popen.return_value = mock_process

        runner = TrainingRunner()
        config = {
            "data_path": "/path/to/data",
            "batch_size": 4,
            "lr": 1e-4,
        }

        result = runner.start_training(config)

        assert result is True
        mock_popen.assert_called_once()
        cmd = mock_popen.call_args[0][0]
        assert cmd[0] == "fluxflow-train"
        assert "--data_path" in cmd
        assert "/path/to/data" in cmd

    @patch("subprocess.Popen")
    def test_sets_is_running_to_true(self, mock_popen):
        """Should set is_running to True after starting."""
        mock_process = MagicMock()
        mock_process.stdout = iter([])
        mock_popen.return_value = mock_process

        runner = TrainingRunner()
        runner.start_training({})

        assert runner.is_running is True

    @patch("subprocess.Popen")
    def test_includes_all_config_options(self, mock_popen):
        """Should include all config options in command."""
        mock_process = MagicMock()
        mock_process.stdout = iter([])
        mock_popen.return_value = mock_process

        runner = TrainingRunner()
        config = {
            "vae_dim": 128,
            "batch_size": 8,
            "use_fp16": True,
            "train_vae": True,
        }

        runner.start_training(config)

        cmd = mock_popen.call_args[0][0]
        assert "--vae_dim" in cmd
        assert "128" in cmd
        assert "--batch_size" in cmd
        assert "8" in cmd
        assert "--use_fp16" in cmd
        assert "--train_vae" in cmd

    @patch("subprocess.Popen")
    def test_returns_false_on_file_not_found(self, mock_popen):
        """Should return False if command not found."""
        mock_popen.side_effect = FileNotFoundError("Command not found")
        runner = TrainingRunner()

        result = runner.start_training({})

        # Should return False due to FileNotFoundError
        assert result is False

    @patch("subprocess.Popen")
    def test_creates_temp_config_for_optimizers(self, mock_popen, tmp_path):
        """Should create temp config file for optimizers/schedulers."""
        mock_process = MagicMock()
        mock_process.stdout = iter([])
        mock_popen.return_value = mock_process

        runner = TrainingRunner()
        config = {
            "output_path": str(tmp_path),
            "optimizers": {"flow": {"type": "AdamW"}},
            "schedulers": {"flow": {"type": "CosineAnnealingLR"}},
        }

        runner.start_training(config)

        assert runner.temp_config_file is not None
        assert runner.temp_config_file.endswith("optim_sched_config.json")


class TestStopTraining:
    """Tests for stopping training."""

    def test_returns_false_if_not_running(self):
        """Should return False if not running."""
        runner = TrainingRunner()

        result = runner.stop_training()

        assert result is False

    @patch("subprocess.Popen")
    def test_terminates_process(self, mock_popen):
        """Should terminate the subprocess."""
        mock_process = MagicMock()
        mock_process.stdout = iter([])
        mock_popen.return_value = mock_process

        runner = TrainingRunner()
        runner.start_training({})

        result = runner.stop_training()

        assert result is True
        mock_process.terminate.assert_called_once()

    @patch("subprocess.Popen")
    def test_sets_is_running_to_false(self, mock_popen):
        """Should set is_running to False after stopping."""
        mock_process = MagicMock()
        mock_process.stdout = iter([])
        mock_popen.return_value = mock_process

        runner = TrainingRunner()
        runner.start_training({})
        runner.stop_training()

        assert runner.is_running is False

    @patch("subprocess.Popen")
    def test_kills_process_on_timeout(self, mock_popen):
        """Should kill process if terminate times out."""
        import subprocess

        mock_process = MagicMock()
        mock_process.stdout = iter([])
        mock_process.wait.side_effect = subprocess.TimeoutExpired("cmd", 5)
        mock_popen.return_value = mock_process

        runner = TrainingRunner()
        runner.start_training({})

        result = runner.stop_training()

        assert result is True
        mock_process.kill.assert_called_once()


class TestGetOutput:
    """Tests for getting training output."""

    def test_returns_empty_list_when_no_output(self):
        """Should return empty list when queue is empty."""
        runner = TrainingRunner()

        output = runner.get_output()

        assert output == []

    def test_returns_all_queued_lines(self):
        """Should return all lines from queue."""
        runner = TrainingRunner()
        runner.output_queue.put("line 1")
        runner.output_queue.put("line 2")
        runner.output_queue.put("line 3")

        output = runner.get_output()

        assert output == ["line 1", "line 2", "line 3"]

    def test_empties_queue_after_get(self):
        """Should empty queue after getting output."""
        runner = TrainingRunner()
        runner.output_queue.put("test")

        runner.get_output()
        output = runner.get_output()

        assert output == []


class TestCleanup:
    """Tests for cleanup functionality."""

    def test_cleanup_removes_temp_file(self, tmp_path):
        """Should remove temp config file on cleanup."""
        runner = TrainingRunner()

        # Create a temp file
        temp_file = tmp_path / "temp_config.json"
        temp_file.write_text("{}")
        runner.temp_config_file = str(temp_file)

        runner._cleanup()

        assert not temp_file.exists()

    def test_cleanup_handles_missing_file(self):
        """Should handle missing temp file gracefully."""
        runner = TrainingRunner()
        runner.temp_config_file = "/nonexistent/file.json"

        # Should not raise
        runner._cleanup()

    def test_cleanup_handles_none_temp_file(self):
        """Should handle None temp file gracefully."""
        runner = TrainingRunner()
        runner.temp_config_file = None

        # Should not raise
        runner._cleanup()


class TestThreadSafety:
    """Tests for thread safety."""

    @patch("subprocess.Popen")
    def test_concurrent_start_calls(self, mock_popen):
        """Should handle concurrent start calls safely."""
        mock_process = MagicMock()
        mock_process.stdout = iter([])
        mock_popen.return_value = mock_process

        runner = TrainingRunner()
        results = []
        lock = threading.Lock()

        def try_start():
            result = runner.start_training({})
            with lock:
                results.append(result)

        threads = [threading.Thread(target=try_start) for _ in range(5)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        # At least one should succeed, most should fail
        # Due to thread timing, exact counts can vary
        assert results.count(True) >= 1
        assert len(results) == 5

    def test_is_running_consistent_across_threads(self):
        """is_running property should be thread-safe to read."""
        runner = TrainingRunner()

        # Manually set internal state to test read consistency
        with runner._lock:
            runner._is_running = True

        values = []
        lock = threading.Lock()

        def check_running():
            for _ in range(100):
                val = runner.is_running
                with lock:
                    values.append(val)

        threads = [threading.Thread(target=check_running) for _ in range(5)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        # All values should be True since we set _is_running manually
        assert all(v is True for v in values)



"""FluxFlow UI - Main application."""

import sys
from pathlib import Path

# Add project root to path
project_root = Path(__file__).parent.parent
sys.path.insert(0, str(project_root))

import gradio as gr  # noqa: E402

from fluxflow_ui.tabs.generation import create_generation_tab  # noqa: E402
from fluxflow_ui.tabs.training import create_training_tab  # noqa: E402
from fluxflow_ui.utils.config_manager import ConfigManager  # noqa: E402
from fluxflow_ui.utils.generation_worker import GenerationWorker  # noqa: E402
from fluxflow_ui.utils.training_runner import TrainingRunner  # noqa: E402


def create_app() -> gr.Blocks:
    """Create FluxFlow UI application.

    Returns:
        Gradio Blocks app
    """
    # Initialize utilities
    training_runner = TrainingRunner()
    generation_worker = GenerationWorker()
    config_manager = ConfigManager()

    # Create app
    app: gr.Blocks
    with gr.Blocks(
        title="FluxFlow UI",
        theme=gr.themes.Soft(),
    ) as app:
        gr.Markdown(
            """
            # 🎨 FluxFlow UI

            Train and generate images with FluxFlow text-to-image models
            """
        )

        with gr.Tabs():
            create_training_tab(training_runner, config_manager)
            create_generation_tab(generation_worker, config_manager)

        gr.Markdown(
            """
            ---
            **FluxFlow** - Flow-based Text-to-Image Generation
            Made with ❤️ by Daniele Camisani
            """
        )

    return app


def main():
    """Main entry point."""
    app = create_app()
    app.launch(
        server_name="0.0.0.0",
        server_port=7860,
        share=False,
        show_error=True,
    )


if __name__ == "__main__":
    main()



"""Background training runner for UI."""

import atexit
import json
import logging
import os
import queue
import subprocess
import threading
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

logger = logging.getLogger(__name__)


class TrainingRunner:
    """Run training in background thread with live output."""

    def __init__(self):
        """Initialize training runner."""
        self._lock = threading.Lock()
        self.process: Optional[subprocess.Popen] = None
        self.output_queue: queue.Queue = queue.Queue()
        self._is_running = False
        self.thread: Optional[threading.Thread] = None
        self.temp_config_file: Optional[str] = None

        # Register cleanup on exit
        atexit.register(self._cleanup)

    @property
    def is_running(self) -> bool:
        """Thread-safe check if training is running."""
        with self._lock:
            return self._is_running

    def _cleanup(self) -> None:
        """Cleanup temp files on exit."""
        if self.temp_config_file and os.path.exists(self.temp_config_file):
            try:
                os.remove(self.temp_config_file)
                logger.debug(f"Cleaned up temp config: {self.temp_config_file}")
            except OSError as e:
                logger.warning(f"Failed to cleanup temp config: {e}")

    def _build_data_args(self, config: Dict[str, Any]) -> List[str]:
        """Build data-related command arguments."""
        cmd = []
        if config.get("data_path"):
            cmd.extend(["--data_path", config["data_path"]])
        if config.get("captions_file"):
            cmd.extend(["--captions_file", config["captions_file"]])

        # WebDataset support (new)
        if config.get("use_webdataset"):
            cmd.append("--use_webdataset")
            if config.get("webdataset_token"):
                cmd.extend(["--webdataset_token", config["webdataset_token"]])
            if config.get("webdataset_url"):
                cmd.extend(["--webdataset_url", config["webdataset_url"]])
            if config.get("webdataset_image_key"):
                cmd.extend(["--webdataset_image_key", config["webdataset_image_key"]])
            if config.get("webdataset_caption_key"):
                cmd.extend(["--webdataset_caption_key", config["webdataset_caption_key"]])

        # Legacy TTI-2M support (deprecated, maps to webdataset)
        elif config.get("use_tt2m"):
            cmd.append("--use_tt2m")
            if config.get("tt2m_token"):
                cmd.extend(["--tt2m_token", config["tt2m_token"]])

        return cmd

    def _build_model_args(self, config: Dict[str, Any]) -> List[str]:
        """Build model-related command arguments."""
        cmd = []
        if config.get("output_path"):
            cmd.extend(["--output_path", config["output_path"]])
        if config.get("model_checkpoint"):
            cmd.extend(["--model_checkpoint", config["model_checkpoint"]])
        if config.get("pretrained_bert_model"):
            cmd.extend(["--pretrained_bert_model", config["pretrained_bert_model"]])
        cmd.extend(
            [
                "--vae_dim",
                str(config.get("vae_dim", 64)),
                "--text_embedding_dim",
                str(config.get("text_embedding_dim", 1024)),
                "--feature_maps_dim",
                str(config.get("feature_maps_dim", 64)),
                "--feature_maps_dim_disc",
                str(config.get("feature_maps_dim_disc", 128)),
            ]
        )
        return cmd

    def _build_training_args(self, config: Dict[str, Any]) -> List[str]:
        """Build training-related command arguments."""
        cmd = [
            "--n_epochs",
            str(config.get("n_epochs", 10)),
            "--batch_size",
            str(config.get("batch_size", 1)),
            "--lr",
            str(config.get("lr", 1e-5)),
            "--lr_min",
            str(config.get("lr_min", 0.1)),
            "--workers",
            str(config.get("workers", 4)),
            "--training_steps",
            str(config.get("training_steps", 1)),
            "--initial_clipping_norm",
            str(config.get("initial_clipping_norm", 1.0)),
        ]
        # Training flags
        if config.get("preserve_lr"):
            cmd.append("--preserve_lr")
        if config.get("use_fp16"):
            cmd.append("--use_fp16")
        if config.get("use_gradient_checkpointing"):
            cmd.append("--use_gradient_checkpointing")

        # Training modes
        for flag in ["train_vae", "train_no_gan", "train_spade", "train_diff", "train_diff_full"]:
            if config.get(flag):
                cmd.append(f"--{flag}")

        # Advanced loss options
        if config.get("use_lpips"):
            cmd.append("--use_lpips")
            cmd.extend(["--lambda_lpips", str(config.get("lambda_lpips", 0.1))])

        # CFG training
        if config.get("cfg_dropout_prob", 0.0) > 0.0:
            cmd.extend(["--cfg_dropout_prob", str(config["cfg_dropout_prob"])])

        # Multi-resolution training
        if config.get("reduced_min_sizes"):
            sizes_str = config["reduced_min_sizes"].strip()
            if sizes_str:
                # Parse comma-separated list
                sizes = [s.strip() for s in sizes_str.split(",")]
                for size in sizes:
                    cmd.extend(["--reduced_min_sizes", size])

        # KL divergence
        cmd.extend(
            [
                "--kl_beta",
                str(config.get("kl_beta", 1.0)),
                "--kl_warmup_steps",
                str(config.get("kl_warmup_steps", 100000)),
                "--kl_free_bits",
                str(config.get("kl_free_bits", 0.0)),
            ]
        )
        return cmd

    def _build_output_args(self, config: Dict[str, Any]) -> List[str]:
        """Build output and logging command arguments."""
        cmd = [
            "--lambda_adv",
            str(config.get("lambda_adv", 0.9)),
            "--log_interval",
            str(config.get("log_interval", 10)),
        ]
        if config.get("no_samples"):
            cmd.append("--no_samples")
        if config.get("test_image_address"):
            for img in config["test_image_address"].strip().split():
                cmd.extend(["--test_image_address", img])
        if config.get("sample_captions"):
            for caption in config["sample_captions"].strip().split("\n"):
                if caption.strip():
                    cmd.extend(["--sample_captions", caption.strip()])
        # Misc
        cmd.extend(
            [
                "--tokenizer_name",
                config.get("tokenizer_name", "distilbert-base-uncased"),
                "--img_size",
                str(config.get("img_size", 1024)),
                "--channels",
                str(config.get("channels", 3)),
            ]
        )
        return cmd

    def _build_optim_sched_args(self, config: Dict[str, Any]) -> List[str]:
        """Build optimizer/scheduler config arguments."""
        cmd = []
        # Option 1: User provides a path to an existing config file
        if config.get("optim_sched_config") and Path(config["optim_sched_config"]).exists():
            cmd.extend(["--optim_sched_config", config["optim_sched_config"]])
        # Option 2: UI provides optimizers/schedulers dicts
        elif "optimizers" in config and "schedulers" in config:
            optim_sched_config = {
                "optimizers": config["optimizers"],
                "schedulers": config["schedulers"],
            }
            output_path = config.get("output_path", "outputs")
            os.makedirs(output_path, exist_ok=True)
            self.temp_config_file = str(Path(output_path) / "optim_sched_config.json")
            with open(self.temp_config_file, "w") as f:
                json.dump(optim_sched_config, f, indent=2)
            cmd.extend(["--optim_sched_config", self.temp_config_file])
        return cmd

    def start_training(
        self, config: Dict[str, Any], on_output: Optional[Callable[[str], None]] = None
    ) -> bool:
        """Start training process in background.

        Args:
            config: Training configuration dictionary (or pipeline YAML path)
            on_output: Callback for each line of output

        Returns:
            True if started successfully, False otherwise
        """
        with self._lock:
            if self._is_running:
                return False

            # Check if using pipeline mode (YAML config)
            if config.get("pipeline_yaml_content"):
                cmd = self._build_pipeline_command(config)
            else:
                # Build command using helper methods (simple mode)
                cmd = ["fluxflow-train"]
                cmd.extend(self._build_data_args(config))
                cmd.extend(self._build_model_args(config))
                cmd.extend(self._build_training_args(config))
                cmd.extend(self._build_output_args(config))
                cmd.extend(self._build_optim_sched_args(config))

            # Start process
            try:
                logger.info(f"Starting training with command: {' '.join(cmd)}")
                self.process = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    universal_newlines=True,
                    bufsize=1,
                )
                self._is_running = True

                # Start output reader thread
                self.thread = threading.Thread(
                    target=self._read_output,
                    args=(on_output,),
                    daemon=True,
                )
                self.thread.start()
                logger.info("Training process started")
                return True

            except FileNotFoundError as e:
                logger.error(f"Training command not found: {e}")
                return False
            except Exception as e:
                logger.error(f"Failed to start training: {e}")
                return False

    def _build_pipeline_command(self, config: Dict[str, Any]) -> List[str]:
        """Build command for pipeline mode training.

        Args:
            config: Config with pipeline_yaml_content

        Returns:
            Command list
        """
        # Save YAML to temp file
        yaml_content = config["pipeline_yaml_content"]
        output_path = config.get("output_path", "outputs")
        os.makedirs(output_path, exist_ok=True)

        self.temp_config_file = str(Path(output_path) / "pipeline_config.yaml")
        with open(self.temp_config_file, "w") as f:
            f.write(yaml_content)

        # Build command with --config flag
        cmd = ["fluxflow-train", "--config", self.temp_config_file]
        return cmd

    def _read_output(self, on_output: Optional[Callable[[str], None]]) -> None:
        """Read process output line by line.

        Args:
            on_output: Callback for each line
        """
        if not self.process or not self.process.stdout:
            return

        for line in self.process.stdout:
            line = line.rstrip()
            self.output_queue.put(line)
            if on_output:
                on_output(line)

        self.process.wait()
        with self._lock:
            self._is_running = False
        logger.info("Training process finished")

    def stop_training(self) -> bool:
        """Stop training process.

        Returns:
            True if stopped successfully
        """
        with self._lock:
            if not self._is_running or not self.process:
                return False

            try:
                self.process.terminate()
                self.process.wait(timeout=5)
                self._is_running = False
                logger.info("Training stopped by user")
                return True
            except subprocess.TimeoutExpired:
                self.process.kill()
                self._is_running = False
                logger.warning("Training process killed after timeout")
                return True
            except Exception as e:
                logger.error(f"Failed to stop training: {e}")
                return False

    def get_output(self) -> List[str]:
        """Get all queued output lines.

        Returns:
            List of output lines
        """
        lines = []
        while not self.output_queue.empty():
            try:
                lines.append(self.output_queue.get_nowait())
            except queue.Empty:
                break
        return lines



"""FluxFlow UI - Flask-based application (Python 3.14 compatible)."""

import base64
import io
import logging
import os
from functools import wraps
from pathlib import Path

from flask import Flask, jsonify, render_template, request
from flask_cors import CORS
from PIL import Image

from fluxflow_ui.utils.config_manager import ConfigManager
from fluxflow_ui.utils.generation_worker import GenerationWorker
from fluxflow_ui.utils.training_runner import TrainingRunner

# Setup logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s | %(levelname)-8s | %(name)s | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger(__name__)

app = Flask(__name__)

# Restrict CORS to localhost only for security
CORS(app, origins=["http://localhost:7860", "http://127.0.0.1:7860"])

# Initialize utilities
training_runner = TrainingRunner()
generation_worker = GenerationWorker()
config_manager = ConfigManager()

# Base directory for file browser (security: prevent path traversal)
FILE_BROWSER_BASE_DIR = os.path.abspath(os.getcwd())


def require_json(f):
    """Decorator to require JSON body in POST requests."""

    @wraps(f)
    def decorated(*args, **kwargs):
        if not request.json:
            return jsonify({"status": "error", "message": "JSON body required"}), 400
        return f(*args, **kwargs)

    return decorated


def safe_path(user_path: str, base_dir: str = FILE_BROWSER_BASE_DIR) -> str:
    """Validate and sanitize path to prevent directory traversal.

    Args:
        user_path: User-provided path
        base_dir: Base directory to restrict access to

    Returns:
        Safe absolute path

    Raises:
        ValueError: If path traversal attempt detected
    """
    # Resolve the full path
    full_path = os.path.realpath(os.path.join(base_dir, user_path))

    # Ensure the path starts with the base directory
    if not full_path.startswith(os.path.realpath(base_dir)):
        raise ValueError("Path traversal attempt detected")

    return full_path


@app.route("/")
def index():
    """Main page."""
    return render_template("index.html")


@app.route("/api/config/training", methods=["GET"])
def get_training_config():
    """Get training configuration."""
    config = config_manager.load_training_config()
    if not config:
        config = config_manager.get_default_training_config()
    return jsonify(config)


@app.route("/api/config/training", methods=["POST"])
@require_json
def save_training_config():
    """Save training configuration."""
    config = request.json
    config_manager.save_training_config(config)
    return jsonify({"status": "success"})


@app.route("/api/config/generation", methods=["GET"])
def get_generation_config():
    """Get generation configuration."""
    config = config_manager.load_generation_config()
    if not config:
        config = config_manager.get_default_generation_config()
    return jsonify(config)


@app.route("/api/config/generation", methods=["POST"])
@require_json
def save_generation_config():
    """Save generation configuration."""
    config = request.json
    config_manager.save_generation_config(config)
    return jsonify({"status": "success"})


@app.route("/api/training/start", methods=["POST"])
@require_json
def start_training():
    """Start training."""
    config = request.json

    if training_runner.is_running:
        return jsonify({"status": "error", "message": "Training already running"}), 400

    success = training_runner.start_training(config)

    if success:
        logger.info("Training started successfully")
        return jsonify({"status": "success", "message": "Training started"})
    else:
        logger.error("Failed to start training")
        return jsonify({"status": "error", "message": "Failed to start training"}), 500


@app.route("/api/training/stop", methods=["POST"])
def stop_training():
    """Stop training."""
    if not training_runner.is_running:
        return jsonify({"status": "error", "message": "No training is running"}), 400

    success = training_runner.stop_training()

    if success:
        logger.info("Training stopped successfully")
        return jsonify({"status": "success", "message": "Training stopped"})
    else:
        logger.error("Failed to stop training")
        return jsonify({"status": "error", "message": "Failed to stop training"}), 500


@app.route("/api/training/status", methods=["GET"])
def training_status():
    """Get training status."""
    return jsonify({"running": training_runner.is_running, "output": training_runner.get_output()})


@app.route("/api/generation/inspect", methods=["POST"])
@require_json
def inspect_model():
    """Inspect model checkpoint and detect dimensions."""
    data = request.json
    checkpoint_path = data.get("checkpoint_path")

    if not checkpoint_path:
        return jsonify({"status": "error", "message": "checkpoint_path required"}), 400

    if not Path(checkpoint_path).exists():
        return jsonify({"status": "error", "message": "Invalid checkpoint path"}), 400

    try:
        import safetensors.torch

        # Load only metadata to detect dimensions
        state_dict = safetensors.torch.load_file(checkpoint_path)

        # Detect VAE dimension from vae_to_dmodel layer (this is the VAE latent dim)
        vae_dim = None
        for key in state_dict.keys():
            if "flow_processor.vae_to_dmodel.weight" in key:
                shape = state_dict[key].shape
                # Shape is [d_model, vae_dim], so vae_dim is shape[1]
                vae_dim = int(shape[1])
                break

        # Detect d_model (feature dimension) from vae_to_dmodel output dimension
        feature_dim = None
        for key in state_dict.keys():
            if "flow_processor.vae_to_dmodel.weight" in key:
                shape = state_dict[key].shape
                # Shape is [d_model, vae_dim], so d_model is shape[0]
                feature_dim = int(shape[0])
                break

        # Fallback to reasonable defaults if detection failed
        if not vae_dim:
            vae_dim = 64
        if not feature_dim:
            feature_dim = 64

        return jsonify(
            {
                "status": "success",
                "vae_dim": int(vae_dim),
                "feature_maps_dim": int(feature_dim),
                "text_embedding_dim": 1024,
                "message": f"Detected: VAE={vae_dim}, Feature={feature_dim}",
            }
        )

    except Exception as e:
        logger.error(f"Failed to inspect model: {e}")
        return jsonify({"status": "error", "message": f"Failed to inspect model: {str(e)}"}), 500


@app.route("/api/generation/load", methods=["POST"])
@require_json
def load_model():
    """Load generation model."""
    data = request.json

    checkpoint_path = data.get("checkpoint_path")
    if not checkpoint_path:
        return jsonify({"status": "error", "message": "checkpoint_path required"}), 400

    success, message = generation_worker.load_model(
        checkpoint_path=checkpoint_path,
        vae_dim=data.get("vae_dim", 64),
        feature_maps_dim=data.get("feature_maps_dim", 64),
        text_embedding_dim=data.get("text_embedding_dim", 1024),
    )

    if success:
        logger.info(f"Model loaded: {checkpoint_path}")
        return jsonify({"status": "success", "message": message})
    else:
        logger.error(f"Failed to load model: {message}")
        return jsonify({"status": "error", "message": message}), 500


@app.route("/api/generation/generate", methods=["POST"])
@require_json
def generate_image():
    """Generate image from prompt."""
    data = request.json

    prompt = data.get("prompt")
    if not prompt:
        return jsonify({"status": "error", "message": "prompt required"}), 400

    image, message = generation_worker.generate_image(
        prompt=prompt,
        img_width=data.get("img_width", 512),
        img_height=data.get("img_height", 512),
        ddim_steps=data.get("ddim_steps", 50),
        seed=data.get("seed") if data.get("use_seed") else None,
        use_cfg=data.get("use_cfg", False),
        guidance_scale=data.get("guidance_scale", 5.0),
        negative_prompt=data.get("negative_prompt"),
    )

    if image is not None:
        # Convert numpy array to base64
        pil_img = Image.fromarray(image)
        buffer = io.BytesIO()
        pil_img.save(buffer, format="PNG")
        buffer.seek(0)
        img_base64 = base64.b64encode(buffer.getvalue()).decode()

        logger.info(f"Image generated for prompt: {prompt[:50]}...")
        return jsonify(
            {
                "status": "success",
                "message": message,
                "image": f"data:image/png;base64,{img_base64}",
            }
        )
    else:
        logger.error(f"Generation failed: {message}")
        return jsonify({"status": "error", "message": message}), 500


@app.route("/api/generation/status", methods=["GET"])
def generation_status():
    """Get generation status."""
    return jsonify({"loaded": generation_worker.is_loaded(), "config": generation_worker.config})


def _format_file_size(size_bytes: int) -> str:
    """Format file size in human-readable format."""
    if size_bytes < 1024:
        return f"{size_bytes}B"
    elif size_bytes < 1024 * 1024:
        return f"{size_bytes / 1024:.1f}KB"
    elif size_bytes < 1024 * 1024 * 1024:
        return f"{size_bytes / (1024 * 1024):.1f}MB"
    else:
        return f"{size_bytes / (1024 * 1024 * 1024):.1f}GB"


def _should_include_file(item: Path, file_type: str) -> bool:
    """Check if file should be included based on filter type."""
    if item.is_dir():
        return True  # Always include directories for navigation
    if file_type == "dir":
        return False  # Skip files when only looking for directories
    if file_type == "safetensors":
        return item.name.endswith(".safetensors")
    return True  # 'all' or 'file' - include everything


def _resolve_browse_path(current_path: str) -> Path:
    """Resolve and validate browse path."""
    if not os.path.isabs(current_path):
        path_obj = Path(current_path).expanduser().resolve()
    else:
        path_obj = Path(current_path).resolve()

    if not path_obj.exists():
        path_obj = Path.cwd()

    if path_obj.is_file():
        path_obj = path_obj.parent

    return path_obj


@app.route("/api/files/browse", methods=["POST"])
@require_json
def browse_files():
    """Browse files and directories with path traversal protection."""
    data = request.json
    current_path = data.get("path", ".")
    file_type = data.get("type", "all")  # 'all', 'dir', 'file', 'safetensors'

    try:
        path_obj = _resolve_browse_path(current_path)
        items = []

        # Add parent directory option
        if path_obj.parent != path_obj:
            items.append({"name": "..", "path": str(path_obj.parent), "type": "dir", "size": ""})

        # List directory contents
        for item in sorted(path_obj.iterdir(), key=lambda x: (not x.is_dir(), x.name.lower())):
            # Skip hidden files
            if item.name.startswith("."):
                continue

            if not _should_include_file(item, file_type):
                continue

            item_type = "dir" if item.is_dir() else "file"
            size = _format_file_size(item.stat().st_size) if item.is_file() else ""

            items.append({"name": item.name, "path": str(item), "type": item_type, "size": size})

        return jsonify({"status": "success", "current_path": str(path_obj), "items": items})

    except PermissionError:
        logger.warning(f"Permission denied accessing path: {current_path}")
        return jsonify({"status": "error", "message": "Permission denied"}), 403
    except Exception as e:
        logger.error(f"Error browsing files: {e}")
        return jsonify({"status": "error", "message": str(e)}), 500


def main():
    """Main entry point for fluxflow-ui command."""
    logger.info("=" * 60)
    logger.info("FluxFlow UI (Flask)")
    logger.info("=" * 60)
    logger.info("Starting server at http://localhost:7860")
    logger.info("Press Ctrl+C to stop")
    logger.info("=" * 60)

    app.run(host="0.0.0.0", port=7860, debug=False, threaded=True)


if __name__ == "__main__":
    main()



"""Tests for ConfigManager."""

import json
import os

from fluxflow_ui.utils.config_manager import ConfigManager


class TestConfigManagerInit:
    """Tests for ConfigManager initialization."""

    def test_creates_config_directory(self, tmp_path):
        """Should create config directory if it doesn't exist."""
        config_dir = tmp_path / "test_configs"
        assert not config_dir.exists()

        ConfigManager(str(config_dir))

        assert config_dir.exists()
        assert config_dir.is_dir()

    def test_sets_config_paths(self, tmp_path):
        """Should set correct paths for config files."""
        config_dir = tmp_path / "configs"
        manager = ConfigManager(str(config_dir))

        assert manager.training_config_path == config_dir / "training_config.json"
        assert manager.generation_config_path == config_dir / "generation_config.json"


class TestSaveTrainingConfig:
    """Tests for saving training configuration."""

    def test_save_valid_config(self, tmp_path):
        """Should save valid config to file."""
        manager = ConfigManager(str(tmp_path / "configs"))
        config = {"batch_size": 4, "lr": 1e-4}

        result = manager.save_training_config(config)

        assert result is True
        assert manager.training_config_path.exists()

        with open(manager.training_config_path) as f:
            saved = json.load(f)
        assert saved == config

    def test_save_complex_config(self, tmp_path):
        """Should save complex nested config."""
        manager = ConfigManager(str(tmp_path / "configs"))
        config = {
            "optimizers": {
                "flow": {"type": "AdamW", "lr": 5e-7},
                "vae": {"type": "AdamW", "lr": 5e-7},
            },
            "schedulers": {
                "flow": {"type": "CosineAnnealingLR"},
            },
        }

        result = manager.save_training_config(config)

        assert result is True
        with open(manager.training_config_path) as f:
            saved = json.load(f)
        assert saved == config

    def test_save_overwrites_existing(self, tmp_path):
        """Should overwrite existing config file."""
        manager = ConfigManager(str(tmp_path / "configs"))

        manager.save_training_config({"version": 1})
        manager.save_training_config({"version": 2})

        with open(manager.training_config_path) as f:
            saved = json.load(f)
        assert saved["version"] == 2

    def test_save_returns_false_on_permission_error(self, tmp_path):
        """Should return False when cannot write file."""
        config_dir = tmp_path / "configs"
        config_dir.mkdir()
        manager = ConfigManager(str(config_dir))

        # Make directory read-only
        os.chmod(config_dir, 0o444)

        try:
            result = manager.save_training_config({"test": True})
            assert result is False
        finally:
            # Restore permissions for cleanup
            os.chmod(config_dir, 0o755)


class TestLoadTrainingConfig:
    """Tests for loading training configuration."""

    def test_load_existing_config(self, tmp_path):
        """Should load existing config file."""
        manager = ConfigManager(str(tmp_path / "configs"))
        config = {"batch_size": 8, "epochs": 10}

        with open(manager.training_config_path, "w") as f:
            json.dump(config, f)

        loaded = manager.load_training_config()

        assert loaded == config

    def test_load_returns_none_when_missing(self, tmp_path):
        """Should return None when config file doesn't exist."""
        manager = ConfigManager(str(tmp_path / "configs"))

        loaded = manager.load_training_config()

        assert loaded is None

    def test_load_returns_none_on_corrupted_json(self, tmp_path):
        """Should return None when config file is corrupted."""
        manager = ConfigManager(str(tmp_path / "configs"))

        with open(manager.training_config_path, "w") as f:
            f.write("not valid json {{{")

        loaded = manager.load_training_config()

        assert loaded is None

    def test_load_returns_none_on_empty_file(self, tmp_path):
        """Should return None when config file is empty."""
        manager = ConfigManager(str(tmp_path / "configs"))

        manager.training_config_path.touch()

        loaded = manager.load_training_config()

        assert loaded is None


class TestSaveGenerationConfig:
    """Tests for saving generation configuration."""

    def test_save_valid_config(self, tmp_path):
        """Should save valid generation config."""
        manager = ConfigManager(str(tmp_path / "configs"))
        config = {"img_size": 512, "ddim_steps": 50}

        result = manager.save_generation_config(config)

        assert result is True
        assert manager.generation_config_path.exists()

    def test_save_returns_false_on_invalid_data(self, tmp_path):
        """Should return False when data cannot be serialized."""
        manager = ConfigManager(str(tmp_path / "configs"))

        # Circular reference cannot be serialized
        config = {}
        config["self"] = config

        result = manager.save_generation_config(config)

        assert result is False


class TestLoadGenerationConfig:
    """Tests for loading generation configuration."""

    def test_load_existing_config(self, tmp_path):
        """Should load existing generation config."""
        manager = ConfigManager(str(tmp_path / "configs"))
        config = {"model_checkpoint": "/path/to/model.safetensors"}

        with open(manager.generation_config_path, "w") as f:
            json.dump(config, f)

        loaded = manager.load_generation_config()

        assert loaded == config

    def test_load_returns_none_when_missing(self, tmp_path):
        """Should return None when config doesn't exist."""
        manager = ConfigManager(str(tmp_path / "configs"))

        loaded = manager.load_generation_config()

        assert loaded is None


class TestDefaultConfigs:
    """Tests for default configuration methods."""

    def test_default_training_config_has_required_keys(self, tmp_path):
        """Default training config should have all required keys."""
        manager = ConfigManager(str(tmp_path / "configs"))

        config = manager.get_default_training_config()

        # Check essential keys
        assert "batch_size" in config
        assert "lr" in config
        assert "n_epochs" in config
        assert "vae_dim" in config
        assert "optimizers" in config
        assert "schedulers" in config

    def test_default_training_config_has_valid_values(self, tmp_path):
        """Default training config should have sensible values."""
        manager = ConfigManager(str(tmp_path / "configs"))

        config = manager.get_default_training_config()

        assert config["batch_size"] > 0
        assert config["lr"] > 0
        assert config["n_epochs"] > 0

    def test_default_generation_config_has_required_keys(self, tmp_path):
        """Default generation config should have all required keys."""
        manager = ConfigManager(str(tmp_path / "configs"))

        config = manager.get_default_generation_config()

        assert "img_size" in config
        assert "ddim_steps" in config
        assert "vae_dim" in config
        assert "model_checkpoint" in config

    def test_default_generation_config_has_valid_values(self, tmp_path):
        """Default generation config should have sensible values."""
        manager = ConfigManager(str(tmp_path / "configs"))

        config = manager.get_default_generation_config()

        assert config["img_size"] > 0
        assert config["ddim_steps"] > 0


class TestRoundTrip:
    """Tests for save/load round-trip consistency."""

    def test_training_config_round_trip(self, tmp_path):
        """Saving and loading training config should preserve data."""
        manager = ConfigManager(str(tmp_path / "configs"))
        original = manager.get_default_training_config()

        manager.save_training_config(original)
        loaded = manager.load_training_config()

        assert loaded == original

    def test_generation_config_round_trip(self, tmp_path):
        """Saving and loading generation config should preserve data."""
        manager = ConfigManager(str(tmp_path / "configs"))
        original = manager.get_default_generation_config()

        manager.save_generation_config(original)
        loaded = manager.load_generation_config()

        assert loaded == original



"""Generation worker for UI."""

import sys
from pathlib import Path
from typing import Optional, Tuple

import numpy as np
import safetensors.torch
import torch
from transformers import AutoTokenizer

# Add project root to path
project_root = Path(__file__).parent.parent.parent
sys.path.insert(0, str(project_root))

from fluxflow.models import (  # noqa: E402
    BertTextEncoder,
    FluxCompressor,
    FluxExpander,
    FluxFlowProcessor,
    FluxPipeline,
)
from fluxflow.utils import generate_latent_images  # noqa: E402


class GenerationWorker:
    """Worker for generating images from text prompts."""

    def __init__(self):
        """Initialize generation worker."""
        self.model_checkpoint: Optional[str] = None
        self.diffuser: Optional[FluxPipeline] = None
        self.text_encoder: Optional[BertTextEncoder] = None
        self.tokenizer: Optional[AutoTokenizer] = None
        self.device = self._get_device()
        self.config = {}

    def _get_device(self) -> torch.device:
        """Get available device.

        Returns:
            torch device
        """
        if torch.cuda.is_available():
            return torch.device("cuda")
        elif torch.backends.mps.is_available():
            return torch.device("mps")
        else:
            return torch.device("cpu")

    def load_model(
        self,
        checkpoint_path: str,
        vae_dim: int = 64,
        feature_maps_dim: int = 64,
        text_embedding_dim: int = 1024,
    ) -> Tuple[bool, str]:
        """Load model from checkpoint.

        Args:
            checkpoint_path: Path to model checkpoint
            vae_dim: VAE latent dimension
            feature_maps_dim: Flow processor dimension
            text_embedding_dim: Text embedding dimension

        Returns:
            Tuple of (success, message)
        """
        try:
            if not Path(checkpoint_path).exists():
                return False, f"Checkpoint not found: {checkpoint_path}"

            # Load tokenizer (uses cache if present, otherwise downloads)
            self.tokenizer = AutoTokenizer.from_pretrained(
                "distilbert-base-uncased", cache_dir="./_cache", local_files_only=False
            )
            if self.tokenizer.pad_token is None:  # type: ignore[union-attr]
                self.tokenizer.pad_token = self.tokenizer.eos_token  # type: ignore[union-attr]
                self.tokenizer.add_special_tokens(  # type: ignore[union-attr]
                    {"pad_token": "[PAD]"}
                )

            # Initialize models
            self.text_encoder = BertTextEncoder(embed_dim=text_embedding_dim)
            self.diffuser = FluxPipeline(
                FluxCompressor(d_model=vae_dim),
                FluxFlowProcessor(d_model=feature_maps_dim, vae_dim=vae_dim),
                FluxExpander(d_model=vae_dim),
            )

            # Load checkpoint
            state_dict = safetensors.torch.load_file(checkpoint_path)
            self.diffuser.load_state_dict(
                {
                    k.replace("diffuser.", ""): v
                    for k, v in state_dict.items()
                    if k.startswith("diffuser.")
                },
                strict=False,
            )
            self.text_encoder.load_state_dict(
                {
                    k.replace("text_encoder.", ""): v
                    for k, v in state_dict.items()
                    if k.startswith("text_encoder.")
                },
                strict=False,
            )

            self.diffuser.to(self.device).eval()
            self.text_encoder.to(self.device).eval()

            self.model_checkpoint = checkpoint_path
            self.config = {
                "vae_dim": vae_dim,
                "feature_maps_dim": feature_maps_dim,
                "text_embedding_dim": text_embedding_dim,
            }

            return True, f"Model loaded successfully on {self.device}"

        except Exception as e:
            return False, f"Failed to load model: {str(e)}"

    def generate_image(
        self,
        prompt: str,
        img_width: int = 512,
        img_height: int = 512,
        ddim_steps: int = 50,
        seed: Optional[int] = None,
        use_cfg: bool = False,
        guidance_scale: float = 5.0,
        negative_prompt: Optional[str] = None,
    ) -> Tuple[Optional[np.ndarray], str]:
        """Generate image from text prompt.

        Args:
            prompt: Text prompt
            img_width: Image width (must be multiple of 16)
            img_height: Image height (must be multiple of 16)
            ddim_steps: Number of diffusion steps
            seed: Random seed (optional)
            use_cfg: Enable classifier-free guidance
            guidance_scale: CFG strength (only used if use_cfg=True)
            negative_prompt: Negative prompt for CFG (optional)

        Returns:
            Tuple of (image array, status message)
        """
        if not self.diffuser or not self.text_encoder or not self.tokenizer:
            return None, "Model not loaded. Please load a checkpoint first."

        try:
            # Validate dimensions are multiples of 16
            if img_width % 16 != 0 or img_height % 16 != 0:
                return (
                    None,
                    f"Width and height must be multiples of 16 (got {img_width}x{img_height})",
                )

            # Set seed if provided
            if seed is not None:
                torch.manual_seed(seed)
                if torch.cuda.is_available():
                    torch.cuda.manual_seed_all(seed)

            with torch.no_grad():
                # Tokenize prompt
                inputs = self.tokenizer(  # type: ignore[operator]
                    prompt,
                    padding="max_length",
                    truncation=True,
                    max_length=512,
                    return_tensors="pt",
                )
                input_ids = inputs["input_ids"].to(self.device)
                attention_mask = inputs["attention_mask"].to(self.device)

                # Encode text
                text_embeddings = self.text_encoder(input_ids, attention_mask=attention_mask)

                # Encode negative prompt if CFG is enabled
                negative_embeddings = None
                if use_cfg and guidance_scale > 1.0:
                    if negative_prompt:
                        neg_inputs = self.tokenizer(  # type: ignore[operator]
                            negative_prompt,
                            padding="max_length",
                            truncation=True,
                            max_length=512,
                            return_tensors="pt",
                        )
                        neg_input_ids = neg_inputs["input_ids"].to(self.device)
                        neg_attention_mask = neg_inputs["attention_mask"].to(self.device)
                        negative_embeddings = self.text_encoder(
                            neg_input_ids, attention_mask=neg_attention_mask
                        )
                    else:
                        # Use null conditioning (zeros)
                        negative_embeddings = torch.zeros_like(text_embeddings)

                # Create random latent with specified dimensions
                z_img = (torch.rand((1, 3, img_height, img_width), device=self.device) * 2) - 1
                latent_z = self.diffuser.compressor(z_img)

                img_seq = latent_z[:, :-1, :].contiguous()
                hw_vec = latent_z[:, -1:, :].contiguous()
                noise_img = torch.randn_like(img_seq)

                # Create noised latent
                from diffusers import DPMSolverMultistepScheduler

                scheduler = DPMSolverMultistepScheduler(num_train_timesteps=1000)
                scheduler.set_timesteps(  # type: ignore[attr-defined]
                    ddim_steps, device=self.device
                )

                t = torch.randint(0, 1000, (1,), device=self.device)
                noised_img = scheduler.add_noise(img_seq, noise_img, t)  # type: ignore
                noised_latent = torch.cat([noised_img, hw_vec], dim=1)

                # Denoise with or without CFG
                if use_cfg and guidance_scale > 1.0 and negative_embeddings is not None:
                    # Use CFG-guided generation
                    denoised_latent = self._generate_with_cfg(
                        noised_latent=noised_latent,
                        text_embeddings=text_embeddings,
                        negative_embeddings=negative_embeddings,
                        guidance_scale=guidance_scale,
                        steps=ddim_steps,
                    )
                else:
                    # Standard generation
                    denoised_latent = generate_latent_images(
                        batch_z=noised_latent,
                        text_embeddings=text_embeddings,
                        diffuser=self.diffuser,
                        steps=ddim_steps,
                        prediction_type="v_prediction",
                    )

                # Decode
                decoded_image = self.diffuser.expander(denoised_latent)

                # Convert to numpy array for Gradio
                image = decoded_image[0].cpu()
                image = (image + 1) / 2  # [-1, 1] -> [0, 1]
                image = image.clamp(0, 1)
                image = image.permute(1, 2, 0).numpy()
                image = (image * 255).astype(np.uint8)

                return image, "Image generated successfully!"

        except Exception as e:
            return None, f"Generation failed: {str(e)}"

    def _generate_with_cfg(
        self,
        noised_latent: torch.Tensor,
        text_embeddings: torch.Tensor,
        negative_embeddings: torch.Tensor,
        guidance_scale: float,
        steps: int,
    ) -> torch.Tensor:
        """Generate image with classifier-free guidance.

        Args:
            noised_latent: Initial noised latent
            text_embeddings: Conditional text embeddings
            negative_embeddings: Unconditional/negative text embeddings
            guidance_scale: CFG strength
            steps: Number of denoising steps

        Returns:
            Denoised latent
        """
        from diffusers import DPMSolverMultistepScheduler

        assert self.diffuser is not None, "Diffuser must be loaded before CFG generation"

        scheduler = DPMSolverMultistepScheduler(num_train_timesteps=1000)
        scheduler.set_timesteps(steps, device=self.device)  # type: ignore

        latent = noised_latent

        for t in scheduler.timesteps:  # type: ignore
            # Expand t to batch dimension
            t_batch = t.unsqueeze(0).to(self.device)

            # Predict with conditional embeddings
            v_cond = self.diffuser.flow_processor(latent, text_embeddings, t_batch)

            # Predict with unconditional embeddings
            v_uncond = self.diffuser.flow_processor(latent, negative_embeddings, t_batch)

            # Apply guidance
            v_guided = v_uncond + guidance_scale * (v_cond - v_uncond)

            # Step the scheduler
            latent = scheduler.step(v_guided, t, latent).prev_sample  # type: ignore

        return latent

    def is_loaded(self) -> bool:
        """Check if model is loaded.

        Returns:
            True if model is loaded
        """
        return self.diffuser is not None and self.text_encoder is not None
